import logging

from tornado import gen
from messenger_sdk.fb_response import FbResponse
from messenger_sdk.templates import Response
from messenger_sdk.fb_manager import FbManager
from messenger_sdk.events import EventFactory

from messenger_sdk.middlewares.middleware import Middleware, EventMiddleware, ResponseMiddleware


class MiddlewareListError(Exception):
    pass


class BaseHandler:
    _custom_middlewares = list()

    def __init__(self, event_factory: EventFactory, fb_manager: FbManager, base_middlewares: list):
        self._event_factory = event_factory
        self._fb_manager = fb_manager
        self._base_middlewares = base_middlewares

    @gen.coroutine
    def handle_entries(self, entries):
        for event in entries:
            try:
                logging.info('Input event: %s', event)
                event = self._event_factory.create_event(event)
                response = FbResponse()
                middleware_classes = self.load_middleware_classes()
                middleware_classes_count = len(middleware_classes)
                for index, middleware_class in enumerate(middleware_classes):
                    if not event.propagate:
                        break
                    yield self._process_middleware(event=event,
                                                   response=response,
                                                   middleware_class=middleware_class)
                    if response.send or index - 1 == middleware_classes_count:
                        yield self._fb_manager.send(response)
                        response = FbResponse()
            except Exception as err:
                logging.error('Unable to handle event: %s', err)
                response = Response(recipient_id=event.recipient_id,
                                    sender_action=Response.SENDER_ACTION_TYPING_OFF).__dict__()
                yield self._fb_manager.send(response)

    @staticmethod
    def _check_middleware_classes_list(middleware_classes):
        if middleware_classes is None:
            raise MiddlewareListError('Middleware classes list cannot be None.')
        if not isinstance(middleware_classes, list):
            raise MiddlewareListError('Middleware classes must be type of list')
        if len(middleware_classes) == 0:
            logging.warning('Empty middleware classes list.')

    def load_middleware_classes(self):
        base_middlewares = self._base_middlewares
        self._check_middleware_classes_list(base_middlewares)

        custom_middlewares = self._custom_middlewares
        self._check_middleware_classes_list(custom_middlewares)

        middlewares = base_middlewares + custom_middlewares

        return sorted(middlewares, key=lambda k: k['priority'], reverse=True)

    def _check_mw_type(self, mw_instance):
        if not isinstance(mw_instance, Middleware):
            raise TypeError('Expected {expected} class, {given} given in {self_name}.'.format(
                expected=Middleware.__class__.__name__,
                given=type(mw_instance).__name__,
                self_name=self.__class__.__name__))

    @gen.coroutine
    def _process_middleware(self, event, response, middleware_class):
        mw_instance = middleware_class.get('class')()
        self._check_mw_type(mw_instance)
        if not any(isinstance(event, supported_event) for supported_event in mw_instance.supported_events):
            return
        logging.debug('<--- Middleware: {middleware_name} --->'.format(middleware_name=mw_instance.__class__.__name__))
        if isinstance(mw_instance, EventMiddleware):
            logging.debug('Executing: {function_name}()'.format(function_name=mw_instance.process_event.__name__))
            yield mw_instance.process_event(event=event)
        if isinstance(mw_instance, ResponseMiddleware):
            logging.debug('Executing: {function_name}()'.format(function_name=mw_instance.process_response.__name__))
            yield mw_instance.process_response(event=event, response=response)



from abc import ABCMeta, abstractmethod
from datetime import datetime
from messenger_sdk.fb_response import FbResponse


class Event:
    __metaclass__ = ABCMeta

    def __init__(self, event):
        self._event = event
        self._propagate = True
        self._created_at = datetime.now()
        self._intent = None
        self._storage = None
        self._response = None

    @property
    def recipient_id(self):
        return self._event.get('sender', {}).get('id')

    @property
    def intent(self):
        return self._intent

    @intent.setter
    def intent(self, intent):
        self._intent = intent

    @property
    def created_at(self):
        return self._created_at

    @abstractmethod
    def payload(self):
        raise NotImplementedError()

    @property
    def storage(self):
        return self._storage

    @storage.setter
    def storage(self, storage: dict):
        self._storage = storage

    @property
    def response(self):
        return self._response

    @response.setter
    def response(self, response: FbResponse):
        if not isinstance(response, FbResponse):
            raise TypeError(
                'Response must be instance of {expected}, {given} given.'.format(expected=FbResponse.__class__.__name__,
                                                                                 given=type(response).__name__))
        self._response = response

    @property
    def propagate(self):
        return self._propagate

    def stop_propagation(self):
        self._propagate = False

    def as_dict(self):
        return {
            'userId': self.recipient_id,
            'createdAt': self.created_at,
            'type': self.__class__.__name__,
            'intent': self.intent,
            'payload': self.payload
        }


class PostbackEvent(Event):
    def __init__(self, event):
        super().__init__(event)

    @property
    def payload(self):
        return self._event.get('postback').get('payload')


class MessageEvent(Event):
    def __init__(self, event):
        super().__init__(event)

    @property
    def payload(self):
        return self._event.get('message').get('text')

    @property
    def attachments(self):
        return self._event.get('message').get('attachments')

    @property
    def is_echo(self):
        return self._event.get('message').get('is_echo')


class QuickReplyMessageEvent(MessageEvent):
    def __init__(self, event):
        super().__init__(event)

    @property
    def payload(self):
        return self._event.get('message').get('quick_reply', {}).get('payload')

    @property
    def text(self):
        return self._event.get('message').get('text')


class DeliveryEvent(Event):
    def __init__(self, event):
        super().__init__(event)

    def payload(self):
        pass


class ReadEvent(Event):
    def __init__(self, event):
        super().__init__(event)

    def payload(self):
        pass


class EventFactory:
    @staticmethod
    def create_event(event):
        if event.get('message'):
            if event.get('message').get('quick_reply'):
                event_class = QuickReplyMessageEvent
            else:
                event_class = MessageEvent
        elif event.get('postback'):
            event_class = PostbackEvent
        elif event.get('delivery'):
            event_class = DeliveryEvent
        elif event.get('read'):
            event_class = ReadEvent
        else:
            raise TypeError('Unsupported event.')

        return event_class(event)



from messenger_sdk.templates import Template


class FbResponse:
    def __init__(self, template=None):
        self._send = False
        self._templates = list()
        if template:
            self.add_template(template)

    def add_template(self, template: Template):
        self._throw_if_not_valid_type(template)
        self._templates.append(template)

    def _throw_if_not_valid_type(self, template):
        if not isinstance(template, Template):
            raise TypeError(
                'Invalid template type. Expected {expected}, given {given}.'.format(
                    expected=Template.__class__.__name__,
                    given=type(template).__name__))

    @property
    def templates(self):
        return self._templates

    @property
    def send(self):
        return self._send

    @send.setter
    def send(self, condition: bool):
        self._send = condition
